from requests import Session

from ..devices.gateway import Gateway
from .property import Property, WrongElementError


class VoltageProperty(Property):
    """
    Object for voltages. It stores the current voltage and the corresponding unit.

    :param gateway: Instance of a Gateway object
    :param session: Instance of a requests.Session object
    :param element_uid: Element UID, something like devolo.VoltageMultiLevelSensor:hdm:ZWave:CBC56091/24
    :param current: Voltage messured at the time of creating this instance
    """

    def __init__(self, gateway: Gateway, session: Session, element_uid: str, current: float):
        if not element_uid.startswith("devolo.VoltageMultiLevelSensor:"):
            raise WrongElementError(f"{element_uid} is not a Voltage Sensor.")

        super().__init__(gateway=gateway, session=session, element_uid=element_uid)
        self.current = current
        self.current_unit = "V"


    def fetch_voltage(self) -> float:
        """
        Update and return the current voltage.

        :return: Voltage value
        """
        response = self.get_data_from_uid_list([self.element_uid])
        self.current = response.get("properties").get("value")
        return self.current



import pytest

from devolo_home_control_api.mydevolo import Mydevolo, WrongUrlError, WrongCredentialsError


class TestMydevolo:
    def test_gateway_ids(self, mydevolo, mock_mydevolo__call):
        assert mydevolo.gateway_ids == [self.gateway.get("id")]

    def test_gateway_ids_empty(self, mydevolo, mock_mydevolo__call):
        with pytest.raises(IndexError):
            mydevolo.gateway_ids

    def test_get_full_url(self, mydevolo, mock_mydevolo__call):
        full_url = mydevolo.get_full_url(self.gateway.get("id"))
        assert full_url == self.gateway.get("full_url")

    def test_get_gateway(self, mydevolo, mock_mydevolo__call):
        details = mydevolo.get_gateway(self.gateway.get("id"))
        assert details.get("gatewayId") == self.gateway.get("id")

    def test_get_gateway_invalid(self, mydevolo, mock_mydevolo__call_raise_WrongUrlError):
        with pytest.raises(WrongUrlError):
            mydevolo.get_gateway(self.gateway.get("id"))

    def test_get_zwave_products(self, mydevolo, mock_mydevolo__call):
        product = mydevolo.get_zwave_products(manufacturer="0x0060", product_type="0x0001", product="0x000")
        assert product.get("name") == "Everspring PIR Sensor SP814"

    def test_get_zwave_products_invalid(self, mydevolo, mock_mydevolo__call_raise_WrongUrlError):
        device_infos = mydevolo.get_zwave_products(manufacturer="0x0070", product_type="0x0001", product="0x000")
        assert len(device_infos) == 0

    def test_maintenance_on(self, mydevolo, mock_mydevolo__call):
        assert not mydevolo.maintenance

    def test_maintenance_off(self, mydevolo, mock_mydevolo__call):
        assert mydevolo.maintenance

    def test_set_password(self, mydevolo):
        mydevolo._gateway_ids = [self.gateway.get("id")]

        mydevolo.password = self.user.get("password")

        assert mydevolo._uuid is None
        assert mydevolo._gateway_ids == []

    def test_set_user(self, mydevolo):
        mydevolo._gateway_ids = [self.gateway.get("id")]

        mydevolo.user = self.user.get("username")

        assert mydevolo._uuid is None
        assert mydevolo._gateway_ids == []

    def test_get_user(self, mydevolo):
        mydevolo.user = self.user.get("username")
        assert mydevolo.user == self.user.get("username")

    def test_get_password(self, mydevolo):
        mydevolo.password = self.user.get("password")
        assert mydevolo.password == self.user.get("password")

    def test_singleton_mydevolo(self):
        with pytest.raises(SyntaxError):
            Mydevolo.get_instance()

        first = Mydevolo()
        with pytest.raises(SyntaxError):
            Mydevolo()

        second = Mydevolo.get_instance()
        assert first is second
        Mydevolo.del_instance()

    def test_uuid(self, mydevolo, mock_mydevolo__call):
        mydevolo._uuid = None
        assert mydevolo.uuid == self.user.get("uuid")

    def test_call_WrongCredentialsError(self, mock_response_wrong_credentials_error):
        mydevolo = Mydevolo()
        with pytest.raises(WrongCredentialsError):
            mydevolo._call("test")
        Mydevolo.del_instance()

    def test_call_WrongUrlError(self, mock_response_wrong_url_error):
        mydevolo = Mydevolo()
        with pytest.raises(WrongUrlError):
            mydevolo._call("test")
        Mydevolo.del_instance()

    def test_call_valid(self, mydevolo, mock_response_valid):
        assert mydevolo._call("test").get("response") == "response"



import pytest

from devolo_home_control_api.properties.property import WrongElementError
from devolo_home_control_api.properties.settings_property import SettingsProperty


@pytest.mark.usefixtures("home_control_instance")
class TestSettingsProperty:
    def test_settings_property_valid(self, gateway_instance, mprm_session):
        setting_property = SettingsProperty(gateway=gateway_instance,
                                            session=mprm_session,
                                            element_uid=f"lis.{self.devices.get('mains').get('uid')}",
                                            led_setting=True,
                                            events_enabled=False,
                                            param_changed=True,
                                            local_switching=False,
                                            remote_switching=True,)
        assert setting_property.led_setting
        assert not setting_property.events_enabled
        assert setting_property.param_changed
        assert not setting_property.local_switching
        assert setting_property.remote_switching

    def test_settings_property_invalid(self, gateway_instance, mprm_session):
        with pytest.raises(WrongElementError):
            SettingsProperty(gateway=gateway_instance, session=mprm_session, element_uid="invalid")

    def test_fetch_general_device_settings_valid(self, mock_mprmrest__extract_data_from_element_uid):
        name, icon, zone_id, events_enabled = \
            self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
                .settings_property.get("general_device_settings").fetch_general_device_settings()
        assert name == self.devices.get('mains').get("properties").get('name')
        assert icon == self.devices.get('mains').get("properties").get('icon')
        assert zone_id == self.devices.get('mains').get("properties").get('zone_id')
        assert events_enabled

    def test_fetch_led_setting_valid(self, mock_mprmrest__extract_data_from_element_uid):
        assert self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .settings_property.get("led").fetch_led_setting()

    def test_fetch_param_changed_valid(self, mock_mprmrest__extract_data_from_element_uid):
        assert not self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .settings_property.get("param_changed").fetch_param_changed_setting()

    def test_fetch_protection_setting_invalid(self):
        with pytest.raises(ValueError):
            self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
                .settings_property.get("protection_setting").fetch_protection_setting(protection_setting="invalid")

    def test_fetch_protection_setting_valid(self, mock_mprmrest__extract_data_from_element_uid):
        local_switch = self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .settings_property.get("protection_setting").fetch_protection_setting(protection_setting="local")
        remote_switch = self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .settings_property.get("protection_setting").fetch_protection_setting(protection_setting="remote")
        assert local_switch
        assert not remote_switch



import json
import sys

import pytest


try:
    with open("test_data.json") as file:
        test_data = json.load(file)
except FileNotFoundError:
    print("Please run tests from within the tests directory.")
    sys.exit(127)


pytest_plugins = ['tests.fixtures.gateway',
                  'tests.fixtures.homecontrol',
                  'tests.fixtures.mprm',
                  'tests.fixtures.mydevolo',
                  'tests.fixtures.properties',
                  'tests.fixtures.publisher',
                  'tests.fixtures.requests',
                  'tests.fixtures.socket']


@pytest.fixture(autouse=True)
def test_data_fixture(request):
    """ Load test data. """
    request.cls.user = test_data.get("user")
    request.cls.gateway = test_data.get("gateway")
    request.cls.devices = test_data.get("devices")


@pytest.fixture()
def fill_device_data(request):
    """ Load test device data. """
    consumption_property = request.cls.homecontrol.devices.get(test_data.get('devices').get("mains").get("uid")) \
        .consumption_property
    consumption_property.get(f"devolo.Meter:{test_data.get('devices').get('mains').get('uid')}").current = 0.58
    consumption_property.get(f"devolo.Meter:{test_data.get('devices').get('mains').get('uid')}").total = 125.68

    binary_switch_property = \
        request.cls.homecontrol.devices.get(test_data.get('devices').get("mains").get("uid")).binary_switch_property
    binary_switch_property.get(f"devolo.BinarySwitch:{test_data.get('devices').get('mains').get('uid')}").state = False

    voltage_property = request.cls.homecontrol.devices.get(test_data.get('devices').get("mains").get("uid")).voltage_property
    voltage_property.get(f"devolo.VoltageMultiLevelSensor:{test_data.get('devices').get('mains').get('uid')}").current = 236



from requests import Session

from ..backend.mprm_rest import MprmRest
from ..devices.gateway import Gateway


class Property(MprmRest):
    """
    Base object for properties. It is not meant to use this directly.

    :param gateway: Instance of a Gateway object
    :param session: Instance of a requests.Session object
    :param element_uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24#2
    """

    def __init__(self, gateway: Gateway, session: Session, element_uid: str):
        self.element_uid = element_uid
        self.device_uid = element_uid.split(":", 1)[1].split("#")[0]
        self.is_online = None
        self._gateway = gateway
        self._session = session
        super().__init__()


class WrongElementError(Exception):
    """ This element was not meant for this property. """



import pytest

import requests

from devolo_home_control_api.devices.zwave import Zwave, get_device_type_from_element_uid,\
    get_device_uid_from_element_uid, get_device_uid_from_setting_uid
from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty

from .mocks.mock_gateway import MockGateway


@pytest.mark.usefixtures("mock_get_zwave_products")
class TestZwave:
    def test_get_property(self, home_control_instance, mock_mprmrest__extract_data_from_element_uid):
        device = Zwave(**self.devices.get("mains").get("properties"))
        gateway = MockGateway(self.gateway.get("id"))
        session = requests.Session()

        device.binary_switch_property = {}
        element_uid = f'devolo.BinarySwitch:{self.devices.get("mains").get("uid")}'
        device.binary_switch_property[element_uid] = \
            BinarySwitchProperty(gateway=gateway,
                                 session=session,
                                 element_uid=element_uid,
                                 state=self.devices.get("mains").get("properties").get("state"))

        assert isinstance(device.get_property("binary_switch")[0], BinarySwitchProperty)

    def test_get_property_invalid(self, mydevolo, mock_mydevolo__call):
        device = Zwave(**self.devices.get("mains").get("properties"))

        with pytest.raises(AttributeError):
            device.get_property("binary_switch")

    def test_battery_level(self, mydevolo, mock_mydevolo__call):
        # TODO: Use battery driven device
        device = Zwave(**self.devices.get("ambiguous_1"))

        assert device.batteryLevel == 55

    def test_device_online_state_state(self, mydevolo, mock_mydevolo__call):
        device = Zwave(**self.devices.get("ambiguous_2"))
        assert device.status == 1

        device = Zwave(**self.devices.get("mains").get("properties"))
        assert device.status == 2

        device = Zwave(**self.devices.get("ambiguous_1"))
        assert device.status not in [1, 2]

    def test_get_device_type_from_element_uid(self):
        assert get_device_type_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2#2") == "devolo.Meter"

    def test_get_device_uid_from_setting_uid(self):
        assert get_device_uid_from_setting_uid("lis.hdm:ZWave:EB5A9F6C/2") == "hdm:ZWave:EB5A9F6C/2"

    def test_get_device_uid_from_setting_uid_secure(self):
        assert get_device_uid_from_setting_uid("lis.hdm:ZWave:EB5A9F6C/2:secure") == "hdm:ZWave:EB5A9F6C/2"

    def test_get_device_uid_from_element_uid(self):
        assert get_device_uid_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2#2") == "hdm:ZWave:F6BF9812/2"

    def test_get_device_uid_from_element_uid_secure(self):
        assert get_device_uid_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2:secure#2") == "hdm:ZWave:F6BF9812/2"

    def test_in_online_online(self, mydevolo):
        device = Zwave(**self.devices.get("mains").get("properties"))
        assert device.is_online()

    # def test_in_online_offline(self, mydevolo):
    #     device = Zwave(**self.devices.get("offline"))
    #     assert not device.is_online()



from datetime import datetime
from typing import Any

from requests import Session

from ..devices.gateway import Gateway
from .property import Property, WrongElementError


class ConsumptionProperty(Property):
    """
    Object for consumptions. It stores the current and total consumption and the corresponding units.

    :param gateway: Instance of a Gateway object
    :param session: Instance of a requests.Session object
    :param element_uid: Element UID, something like devolo.Meter:hdm:ZWave:CBC56091/24#2
    :param **kwargs: Consumption values valid at time of creating the instance
    """

    def __init__(self, gateway: Gateway, session: Session, element_uid: str, **kwargs: Any):
        if not element_uid.startswith("devolo.Meter:"):
            raise WrongElementError(f"{element_uid} is not a Meter.")

        super().__init__(gateway=gateway, session=session, element_uid=element_uid)
        self.current = kwargs.get("current")
        self.current_unit = "W"
        self.total = kwargs.get("total")
        self.total_since = datetime.fromtimestamp(kwargs.get("total_since") / 1000)
        self.total_unit = "kWh"


    def fetch_consumption(self, consumption_type: str = "current") -> float:
        """
        Update and return the consumption, specified in consumption_type for the given uid.

        :param consumption_type: Current or total consumption
        :return: Consumption
        """
        if consumption_type not in ["current", "total"]:
            raise ValueError('Unknown consumption type. "current" and "total" are valid consumption types.')
        response = self.get_data_from_uid_list([self.element_uid])
        if consumption_type == "current":
            self.current = response.get("properties").get("currentValue")
            return self.current
        else:
            self.total = response.get("properties").get("totalValue")
            return self.total



import json
import logging

from requests import ReadTimeout

from ..mydevolo import Mydevolo


class MprmRest:
    """
    The abstract MprmRest object handles calls to the so called mPRM. It does not cover all API calls, just those requested
    up to now. All calls are done in a gateway context, so you have to create a derived class, that provides a Gateway object
    and a Session object.
    """

    def __init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._mydevolo = Mydevolo.get_instance()
        self._data_id = 0
        self._local_ip = None


    @property
    def all_devices(self) -> list:
        """
        Get all devices.

        :return: All devices and their properties.
        """
        self._logger.info("Inspecting devices")
        data = {"method": "FIM/getFunctionalItems",
                "params": [['devolo.DevicesPage'], 0]}
        response = self.post(data)
        return response.get("result").get("items")[0].get("properties").get("deviceUIDs")


    def get_data_from_uid_list(self, uids: list) -> list:
        """
        Returns data from an element UID list using an RPC call.

        :param uids: Element UIDs, something like [devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2,
                     devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#1]
        :return: Data connected to the element UIDs, payload so to say
        """
        data = {"method": "FIM/getFunctionalItems",
                "params": [uids, 0]}
        response = self.post(data)
        return response.get("result").get("items")

    def get_name_and_element_uids(self, uid: str):
        """
        Returns the name, all element UIDs and the device model of the given device UID.

        :param uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
        """
        data = {"method": "FIM/getFunctionalItems",
                "params": [[uid], 0]}
        response = self.post(data)
        properties = response.get("result").get("items")[0].get("properties")
        return properties

    def post(self, data: dict) -> dict:
        """
        Communicate with the RPC interface.

        :param data: Data to be send
        :return: Response to the data
        """
        self._data_id += 1
        data['jsonrpc'] = "2.0"
        data['id'] = self._data_id
        try:
            response = self._session.post(self._session.url + "/remote/json-rpc",
                                          data=json.dumps(data),
                                          headers={"content-type": "application/json"},
                                          timeout=30).json()
        except ReadTimeout:
            self._logger.error("Gateway is offline.")
            self._gateway.update_state(False)
            raise MprmDeviceCommunicationError("Gateway is offline.") from None
        if response['id'] != data['id']:
            self._logger.error("Got an unexpected response after posting data.")
            raise ValueError("Got an unexpected response after posting data.")
        return response


class MprmDeviceCommunicationError(Exception):
    """ Communicating to a device via mPRM failed """


class MprmDeviceNotFoundError(Exception):
    """ A device like this was not found """



import socket
import time
from json import JSONDecodeError
from threading import Thread

import requests
from zeroconf import DNSRecord, ServiceBrowser, ServiceStateChange, Zeroconf

from .mprm_rest import MprmDeviceCommunicationError
from .mprm_websocket import MprmWebsocket


class Mprm(MprmWebsocket):
    """
    The abstract Mprm object handles the connection to the devolo Cloud (remote) or the gateway in your LAN (local). Either
    way is chosen, depending on detecting the gateway via mDNS.
    """

    def __init__(self):
        super().__init__()
        self.detect_gateway_in_lan()
        self.create_connection()


    def create_connection(self):
        """ Create session, either locally or via cloud. """
        if self._local_ip:
            self._gateway.local_connection = True
            self.get_local_session()
        elif self._gateway.external_access and not self._mydevolo.maintenance:
            self.get_remote_session()
        else:
            self._logger.error("Cannot connect to gateway. No gateway found in LAN and external access is not possible.")
            raise ConnectionError("Cannot connect to gateway.")

    def detect_gateway_in_lan(self):
        """ Detects a gateway in local network and check if it is the desired one. """
        zeroconf = Zeroconf()
        browser = ServiceBrowser(zeroconf, "_http._tcp.local.", handlers=[self._on_service_state_change])
        self._logger.info("Searching for gateway in LAN")
        start_time = time.time()
        while not time.time() > start_time + 3 and self._local_ip is None:
            for mdns_name in zeroconf.cache.entries():
                self._try_local_connection(mdns_name)
            else:
                time.sleep(0.05)
        Thread(target=browser.cancel).start()
        Thread(target=zeroconf.close).start()
        return self._local_ip

    def get_local_session(self):
        """ Connect to the gateway locally. """
        self._logger.info("Connecting to gateway locally")
        self._session.url = "http://" + self._local_ip
        try:
            self._token_url = self._session.get(self._session.url + "/dhlp/portal/full",
                                                auth=(self._gateway.local_user, self._gateway.local_passkey), timeout=5).json()
        except JSONDecodeError:
            self._logger.error("Could not connect to the gateway locally.")
            raise MprmDeviceCommunicationError("Could not connect to the gateway locally.") from None
        except requests.ConnectTimeout:
            self._logger.error("Timeout during connecting to the gateway.")
            raise
        self._session.get(self._token_url.get('link'))

    def get_remote_session(self):
        """ Connect to the gateway remotely. """
        self._logger.info("Connecting to gateway via cloud")
        try:
            self._session.get(self._gateway.full_url, timeout=15)
        except JSONDecodeError:
            raise MprmDeviceCommunicationError("Gateway is offline.") from None


    def _on_service_state_change(self, zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange):
        """ Service handler for Zeroconf state changes. """
        if state_change is ServiceStateChange.Added:
            zeroconf.get_service_info(service_type, name)

    def _try_local_connection(self, mdns_name: DNSRecord):
        """ Try to connect to an MDNS hostname. If connection was successful, save local IP. """
        try:
            ip = socket.inet_ntoa(mdns_name.address)
            if mdns_name.key.startswith("devolo-homecontrol") and \
                requests.get("http://" + ip + "/dhlp/port/full",
                             auth=(self._gateway.local_user, self._gateway.local_passkey),
                             timeout=0.5).status_code == requests.codes.ok:
                self._logger.debug(f"Got successful answer from ip {ip}. Setting this as local gateway")
                self._local_ip = ip
        except (OSError, AttributeError):
            # OSError: Got IPv6 address which isn't supported by socket.inet_ntoa and the gateway as well.
            # AttributeError: The MDNS entry does not provide address information
            # TODO: We can somehow delete the ip in zeroconf, because if we can't connect once, we won't connect at second try
            pass



import logging

from ..mydevolo import Mydevolo


class Gateway:
    """
    Representing object for devolo Home Control Central Units. As it is a gateway from the IP world to the Z-Wave
    world, we call it that way.

    :param gateway_id: Gateway ID (aka serial number), typically found on the label of the device
    """

    def __init__(self, gateway_id: str):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._mydevolo = Mydevolo.get_instance()
        self._full_url = None

        details = self._mydevolo.get_gateway(gateway_id)

        self.id = details.get("gatewayId")
        self.name = details.get("name")
        self.role = details.get("role")
        self.local_user = self._mydevolo.uuid
        self.local_passkey = details.get("localPasskey")
        self.local_connection = False
        self.external_access = details.get("externalAccess")
        self.firmware_version = details.get("firmwareVersion")

        self._update_state(status=details.get("status"), state=details.get("state"))


    @property
    def full_url(self):
        """ The full URL is used to get a valid remote session. """
        if self._full_url is None:
            self._full_url = self._mydevolo.get_full_url(self.id)
            self._logger.debug(f"Setting full URL to {self._full_url}")
        return self._full_url


    def update_state(self, online: bool = None):
        """
        Update the state of the gateway. If called without parameter, we will check my devolo.

        :param online: Detected state of the gateway
        """
        if online is None:
            details = self._mydevolo.get_gateway(self.id)
            self._update_state(status=details.get("status"), state=details.get("state"))
        else:
            self.online = online
            self.sync = online


    def _update_state(self, status: str, state: str):
        """ Helper to update the state. """
        if status == "devolo.hc_gateway.status.online":
            self.online = True
        else:
            self.online = False
        if state == "devolo.hc_gateway.state.idle":
            self.sync = True
        else:
            self.sync = False



import logging

import requests


class Mydevolo:
    """
    The Mydevolo object handles calls to the my devolo API v1 as singleton. It does not cover all API calls, just
    those requested up to now. All calls are done in a user context, so you need to provide credentials of that user.

    We differentiate between general information like UUID or gateway IDs and information my devolo can provide, if
    you know what you are looking for like gateway details. We treat the former as properties and the latter as
    parametrised functions. Although they typically start with get, those are not getter function, as the result is
    not stored in the object.
    """

    __instance = None

    @classmethod
    def get_instance(cls):
        if cls.__instance is None:
            raise SyntaxError(f"Please init {cls.__name__}() once to establish a connection to my devolo.")
        return cls.__instance

    @classmethod
    def del_instance(cls):
        cls.__instance = None


    def __init__(self):
        if self.__class__.__instance is not None:
            raise SyntaxError(f"Please use {self.__class__.__name__}.get_instance() and reuse the connection to my devolo.")

        self._logger = logging.getLogger(self.__class__.__name__)
        self._user = None
        self._password = None
        self._uuid = None
        self._gateway_ids = []

        self.url = "https://www.mydevolo.com"

        self.__class__.__instance = self


    @property
    def user(self) -> str:
        """ The user (also known as my devolo ID) is used for basic authentication. """
        return self._user

    @user.setter
    def user(self, user: str):
        """ Invalidate uuid and gateway IDs on user name change. """
        self._user = user
        self._uuid = None
        self._gateway_ids = []

    @property
    def password(self) -> str:
        """ The password is used for basic authentication. """
        return self._password

    @password.setter
    def password(self, password: str):
        """ Invalidate uuid and gateway IDs on password change. """
        self._password = password
        self._uuid = None
        self._gateway_ids = []

    @property
    def uuid(self) -> str:
        """ The uuid is a central attribute in my devolo. Most URLs in the user context contain it. """
        if self._uuid is None:
            self._logger.debug("Getting UUID")
            self._uuid = self._call(f"{self.url}/v1/users/uuid").get("uuid")
        return self._uuid

    @property
    def maintenance(self) -> bool:
        """ If devolo Home Control is in maintenance, there is not much we can do via cloud. """
        state = self._call(f"{self.url}/v1/hc/maintenance").get("state")
        if state == "on":
            return False
        else:
            self._logger.warning("devolo Home Control is in maintenance mode.")
            return True

    @property
    def gateway_ids(self) -> list:
        """ Get gateway IDs. """
        if not self._gateway_ids:
            self._logger.debug(f"Getting list of gateways")
            items = self._call(f"{self.url}/v1/users/{self.uuid}/hc/gateways/status").get("items")
            for gateway in items:
                self._gateway_ids.append(gateway.get("gatewayId"))
                self._logger.debug(f'Adding {gateway.get("gatewayId")} to list of gateways.')
            if len(self._gateway_ids) == 0:
                self._logger.error("Could not get gateway list. No Gateway attached to account?")
                raise IndexError("No gateways")
        return self._gateway_ids


    def get_gateway(self, gateway_id: str) -> dict:
        """
        Get gateway details like name, local passkey and other.

        :param gateway_id: Gateway ID
        :return: Gateway object
        """
        self._logger.debug(f"Getting details for gateway {gateway_id}")
        details = {}
        try:
            details = self._call(f"{self.url}/v1/users/{self.uuid}/hc/gateways/{gateway_id}")
        except WrongUrlError:
            self._logger.error("Could not get full URL. Wrong gateway ID used?")
            raise
        return details

    def get_full_url(self, gateway_id: str) -> str:
        """
        Get gateway's portal URL.

        :param gateway_id: Gateway ID
        :return: URL
        """
        self._logger.debug("Getting full URL of gateway.")
        return self._call(f"{self.url}/v1/users/{self.uuid}/hc/gateways/{gateway_id}/fullURL").get("url")

    def get_zwave_products(self, manufacturer: str, product_type: str, product: str) -> dict:
        """
        Get information about a Z-Wave device.

        :param manufacturer: The manufacturer ID in hex.
        :param product_type: The product type ID in hex.
        :param product: The product ID in hex.
        :return: All known product information.
        """
        self._logger.debug(f"Getting information for {manufacturer}/{product_type}/{product}")
        device_info = {}
        try:
            device_info = self._call(f"{self.url}/v1/zwave/products/{manufacturer}/{product_type}/{product}")
        except WrongUrlError:
            # At some devices no device information are returned
            self._logger.debug("No device info found")
        return device_info


    def _call(self, url: str) -> dict:
        """ Make a call to any entry point with the user's context. """
        responds = requests.get(url,
                                auth=(self._user, self._password),
                                headers={'content-type': 'application/json'},
                                timeout=60)
        if responds.status_code == requests.codes.forbidden:
            self._logger.error("Could not get full URL. Wrong username or password?")
            raise WrongCredentialsError("Wrong username or password.")
        if responds.status_code == requests.codes.not_found:
            raise WrongUrlError(f"Wrong URL: {url}")
        return responds.json()


class WrongCredentialsError(Exception):
    """ Wrong credentials were used. """


class WrongUrlError(Exception):
    """ Wrong URL was used. """



from typing import Callable


class Publisher:
    """
    The Publisher send messages to attached subscribers.
    """

    def __init__(self, events: list):
        self._events = {event: dict() for event in events}


    def dispatch(self, event: str, message: tuple):
        """ Dispatch the message to the subscribers. """
        for callback in self._get_subscribers_for_specific_event(event).values():
            callback(message)

    def register(self, event: str, who: object, callback: Callable = None):
        """
        As a new subscriber for an event, add a callback function to call on new message.
        If no callback is given, it registers update().

        :raises AttributeError: The supposed callback is not callable.
        """
        if callback is None:
            callback = getattr(who, 'update')
        self._get_subscribers_for_specific_event(event)[who] = callback

    def unregister(self, event: str, who: object):
        """ Remove a subscriber for a specific event. """
        del self._get_subscribers_for_specific_event(event)[who]


    def _get_subscribers_for_specific_event(self, event: str):
        """ All subscribers listening to an event. """
        return self._events[event]



from requests import Session

from ..devices.gateway import Gateway
from .property import Property, WrongElementError


class BinarySwitchProperty(Property):
    """
    Object for binary switches. It stores the binary switch state.

    :param gateway: Instance of a Gateway object
    :param session: Instance of a requests.Session object
    :param element_uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24#2
    :param state: State the switch has at time of creating this instance
    """

    def __init__(self, gateway: Gateway, session: Session, element_uid: str, state: bool):
        if not element_uid.startswith("devolo.BinarySwitch:"):
            raise WrongElementError(f"{element_uid} is not a Binary Switch.")

        super().__init__(gateway=gateway, session=session, element_uid=element_uid)
        self.state = state


    def fetch_binary_switch_state(self) -> bool:
        """
        Update and return the binary switch state for the given uid.

        :return: Binary switch state
        """
        response = self.get_data_from_uid_list([self.element_uid])
        self.state = True if response.get("properties").get("state") == 1 else False
        return self.state

    def set_binary_switch(self, state: bool):
        """
        Set the binary switch of the given element_uid to the given state.

        :param state: True if switching on, False if switching off
        """
        data = {"method": "FIM/invokeOperation",
                "params": [self.element_uid, "turnOn" if state else "turnOff", []]}
        response = self.post(data)
        if response.get("result").get("status") == 1:
            self.state = state



import pytest

from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty
from devolo_home_control_api.properties.property import WrongElementError


@pytest.mark.usefixtures("home_control_instance")
class TestBinarySwitchProperty:
    def test_binary_switch_property_invalid(self, gateway_instance, mprm_session):
        with pytest.raises(WrongElementError):
            BinarySwitchProperty(gateway=gateway_instance, session=mprm_session, element_uid="invalid", state=True)

    def test_fetch_binary_switch_state_valid_on(self, mock_mprmrest__extract_data_from_element_uid):
        assert self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .binary_switch_property.get(self.devices.get("mains").get("elementUIDs")[1]).fetch_binary_switch_state()

    def test_fetch_binary_switch_state_valid_off(self, mock_mprmrest__extract_data_from_element_uid):
        assert not self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .binary_switch_property.get(self.devices.get("mains").get("elementUIDs")[1]).fetch_binary_switch_state()

    def test_set_binary_switch_valid(self, mock_mprmrest__post_set):
        self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .binary_switch_property.get(self.devices.get("mains").get("elementUIDs")[1]).set_binary_switch(True)
        assert self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .binary_switch_property.get(self.devices.get("mains").get("elementUIDs")[1]).state



import pytest

from devolo_home_control_api.devices.gateway import Gateway


@pytest.mark.usefixtures("mydevolo")
@pytest.mark.usefixtures("mock_mydevolo__call")
class TestGateway:
    def test_update_state_known(self):
        gateway = Gateway(self.gateway.get("id"))
        gateway.update_state(False)

        assert not gateway.online
        assert not gateway.sync

    def test_update_state_offline(self):
        gateway = Gateway(self.gateway.get("id"))
        gateway._update_state(status="devolo.hc_gateway.status.offline", state="devolo.hc_gateway.state.update")

        assert not gateway.online
        assert not gateway.sync

    def test_update_state_unknown(self):
        gateway = Gateway(self.gateway.get("id"))
        gateway.online = False
        gateway.sync = False
        gateway.update_state()

        assert gateway.online
        assert gateway.sync

    @pytest.mark.usefixtures("mock_mydevolo_full_url")
    def test_full_url(self):
        gateway = Gateway(self.gateway.get("id"))
        assert gateway.full_url == self.gateway.get("id")



import pytest
from requests import ConnectTimeout

from devolo_home_control_api.backend.mprm_rest import MprmDeviceCommunicationError

from .mocks.mock_dnsrecord import MockDNSRecord


@pytest.mark.usefixtures("mprm_instance")
class TestMprm:
    @pytest.mark.usefixtures("mock_mprm_get_local_session")
    def test_create_connection_local(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway.get("local_ip")
        self.mprm.create_connection()

    @pytest.mark.usefixtures("mock_mprmwebsocket_get_remote_session")
    @pytest.mark.usefixtures("mock_session_get")
    def test_create_connection_remote(self, mprm_session, mydevolo):
        self.mprm._session = mprm_session
        self.mprm._gateway.external_access = True
        self._mydevolo = mydevolo
        self.mprm.create_connection()

    def test_create_connection_invalid(self):
        with pytest.raises(ConnectionError):
            self.mprm._gateway.external_access = False
            self.mprm.create_connection()

    @pytest.mark.usefixtures("mock_mprm__try_local_connection")
    @pytest.mark.usefixtures("mock_mprm_zeroconf_cache_entries")
    def test_detect_gateway_in_lan(self):
        assert self.mprm.detect_gateway_in_lan() == self.gateway.get("local_ip")

    @pytest.mark.usefixtures("mock_session_get")
    @pytest.mark.usefixtures("mock_response_json")
    def test_get_local_session_valid(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway.get("local_ip")
        self.mprm.get_local_session()

    @pytest.mark.usefixtures("mock_response_requests_ConnectTimeout")
    def test_get_local_session_ConnectTimeout(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway.get("local_ip")
        with pytest.raises(ConnectTimeout):
            self.mprm.get_local_session()

    @pytest.mark.usefixtures("mock_response_json_JSONDecodeError")
    def test_get_local_session_JSONDecodeError(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway.get("local_ip")
        with pytest.raises(MprmDeviceCommunicationError):
            self.mprm.get_local_session()

    @pytest.mark.usefixtures("mock_response_json_JSONDecodeError")
    def test_get_remote_session_JSONDecodeError(self, mprm_session):
        self.mprm._session = mprm_session
        with pytest.raises(MprmDeviceCommunicationError):
            self.mprm.get_remote_session()

    @pytest.mark.usefixtures("mock_socket_inet_ntoa")
    @pytest.mark.usefixtures("mock_response_valid")
    def test__try_local_connection_success(self, mprm_session):
        mdns_name = MockDNSRecord()
        mdns_name.address = self.gateway.get("local_ip")
        self.mprm._session = mprm_session
        self.mprm._try_local_connection(mdns_name)
        assert self.mprm._local_ip == self.gateway.get("local_ip")



import json
import threading
import time

import websocket
from requests import ConnectionError, ReadTimeout
from urllib3.connection import ConnectTimeoutError

from .mprm_rest import MprmDeviceCommunicationError, MprmRest


class MprmWebsocket(MprmRest):
    """
    The abstract MprmWebsocket object handles calls to the mPRM via websockets. It does not cover all API calls, just those
    requested up to now. All calls are done in a gateway context, so you have to create a derived class, that provides a
    Gateway object and a Session object. Further, the derived class needs to implement methods to connect to the websocket,
    either local or remote. Last but not least, the derived class needs to implement a method that is called on new messages.

    The websocket connection itself runs in a thread, that might not terminate as expected. Using a with-statement is
    recommended.
    """

    def __init__(self):
        super().__init__()
        self._ws = None
        self._event_sequence = 0

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.websocket_disconnect()


    def get_local_session(self):
        raise NotImplementedError(f"{self.__class__.__name__} needs a method to connect locally to a gateway.")

    def get_remote_session(self):
        raise NotImplementedError(f"{self.__class__.__name__} needs a method to connect remotely to a gateway.")

    def on_update(self, message):
        raise NotImplementedError(f"{self.__class__.__name__} needs a method to process messages from the websocket.")

    def websocket_connect(self):
        """ Set up the websocket connection. """
        ws_url = self._session.url.replace("https://", "wss://").replace("http://", "ws://")
        cookie = "; ".join([str(name) + "=" + str(value) for name, value in self._session.cookies.items()])
        ws_url = f"{ws_url}/remote/events/?topics=com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED," \
                 f"com/prosyst/mbs/services/fim/FunctionalItemEvent/UNREGISTERED" \
                 f"&filter=(|(GW_ID={self._gateway.id})(!(GW_ID=*)))"
        self._logger.debug(f"Connecting to {ws_url}")
        self._ws = websocket.WebSocketApp(ws_url,
                                          cookie=cookie,
                                          on_open=self._on_open,
                                          on_message=self._on_message,
                                          on_error=self._on_error,
                                          on_close=self._on_close)
        self._ws.run_forever(ping_interval=30, ping_timeout=5)

    def websocket_disconnect(self, event: str = ""):
        """ Close the websocket connection. """
        self._logger.info("Closing web socket connection.")
        if event != "":
            self._logger.info(f"Reason: {event}")
        self._ws.close()


    def _on_close(self):
        """ Callback function to react on closing the websocket. """
        self._logger.info("Closed web socket connection.")

    def _on_error(self, error: str):
        """ Callback function to react on errors. We will try reconnecting with prolonging intervals. """
        self._logger.error(error)
        self.connected = False
        self._ws.close()
        self._event_sequence = 0

        sleep_interval = 16
        while not self.connected:
            self._try_reconnect(sleep_interval)
            sleep_interval = sleep_interval * 2 if sleep_interval < 2048 else 3600

        self.websocket_connect()

    def _on_message(self, message: str):
        """ Callback function to react on a message. """
        message = json.loads(message)

        event_sequence = message.get("properties").get("com.prosyst.mbs.services.remote.event.sequence.number")
        if event_sequence == self._event_sequence:
            self._event_sequence += 1
        else:
            self._logger.warning("We missed a websocket message.")
            self._event_sequence = event_sequence

        self.on_update(message)

    def _on_open(self):
        """ Callback function to keep the websocket open. """
        def run(*args):
            self._logger.info("Starting web socket connection.")
            while self._ws.sock is not None and self._ws.sock.connected:
                time.sleep(1)
        threading.Thread(target=run).start()

    def _try_reconnect(self, sleep_interval: int):
        """ Try to reconnect to the websocket. """
        try:
            self._logger.info("Trying to reconnect to the websocket.")
            # TODO: Check if local_ip is still correct after lost connection
            self.get_local_session() if self._local_ip else self.get_remote_session()
            self.connected = True
        except (json.JSONDecodeError, ConnectTimeoutError, ReadTimeout, ConnectionError, MprmDeviceCommunicationError):
            self._logger.info(f"Sleeping for {sleep_interval} seconds.")
            time.sleep(sleep_interval)



import json
import logging
from datetime import datetime

from .publisher import Publisher
from ..devices.gateway import Gateway
from ..devices.zwave import get_device_type_from_element_uid, get_device_uid_from_element_uid


class Updater:
    """
    The Updater takes care of new states and values of devices and sends them to the Publisher object. Using methods in here
    do not effect the real device states.

    :param devices: List of devices to await updates for
    :param gateway: Instance of a Gateway object
    :param publisher: Instance of a Publisher object
    """

    def __init__(self, devices: dict, gateway: Gateway, publisher: Publisher):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._gateway = gateway
        self._publisher = publisher

        self.devices = devices
        self.on_device_change = None


    def update(self, message: dict):
        """
        Update states and values depending on the message type.

        :param message: Message to process
        """
        message_type = {"devolo.BinarySwitch": self._binary_switch,
                        "devolo.mprm.gw.GatewayAccessibilityFI": self._gateway_accessible,
                        "devolo.Meter": self._meter,
                        "devolo.VoltageMultiLevelSensor": self._voltage_multi_level_sensor,
                        "hdm": self._device_online_state,
                        "devolo.DevicesPage": self._device_change,
                        "devolo.DeviceEvents": self._device_events}
        try:
            message_type[get_device_type_from_element_uid(message.get("properties").get("uid"))](message)
        except KeyError:
            self._logger.debug(json.dumps(message, indent=4))

    def update_device_online_state(self, device_uid: str, value: int):
        """
        Update the device's online state. The value is written into the internal dict.

        :param device_uid: Device UID, something like hdm:ZWave:CBC56091/24#2
        :param value: 1 for online, all other numbers are treated as offline
        """
        self._logger.debug(f"Updating device online state of {device_uid} to {value}")
        self.devices.get(device_uid).status = value
        self._publisher.dispatch(device_uid, (device_uid, value))

    def update_binary_switch_state(self, element_uid: str, value: bool):
        """
        Update the binary switch state of a device externally. The value is written into the internal dict.

        :param element_uid: Element UID, something like, devolo.BinarySwitch:hdm:ZWave:CBC56091/24#2
        :param value: True for on, False for off
        """
        if element_uid.split(".")[-2] == "smartGroup":
            # We ignore if a group is switched. We get the information separated for every device.
            return
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices.get(device_uid).binary_switch_property.get(element_uid).state = value
        self._logger.debug(f"Updating state of {element_uid} to {value}")
        self._publisher.dispatch(device_uid, (element_uid, value))

    def update_consumption(self, element_uid: str, consumption: str, value: float):
        """
        Update the consumption of a device externally. The value is written into the internal dict.

        :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
        :param consumption: current or total consumption
        :param value: Value so be set
        """
        device_uid = get_device_uid_from_element_uid(element_uid)
        if consumption == "current":
            self.devices.get(device_uid).consumption_property.get(element_uid).current = value
        else:
            self.devices.get(device_uid).consumption_property.get(element_uid).total = value
        self._logger.debug(f"Updating {consumption} consumption of {element_uid} to {value}")
        self._publisher.dispatch(device_uid, (element_uid, value))

    def update_voltage(self, element_uid: str, value: float):
        """
        Update the voltage of a device externally. The value is written into the internal dict.

        :param element_uid: Element UID, something like devolo.VoltageMultiLevelSensor:hdm:ZWave:CBC56091/24
        :param value: Value so be set
        """
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices.get(device_uid).voltage_property.get(element_uid).current = value
        self._logger.debug(f"Updating voltage of {element_uid} to {value}")
        self._publisher.dispatch(device_uid, (element_uid, value))

    def update_gateway_state(self, accessible: bool, online_sync: bool):
        """
        Update the gateway status externally. A gateway might go on- or offline while we listen to the websocket.

        :param accessible: Online state of the gateway
        :param online_sync: Sync state of the gateway
        """
        self._logger.debug(f"Updating status and state of gateway to status: {accessible} and state: {online_sync}")
        self._gateway.online = accessible
        self._gateway.sync = online_sync

    def update_total_since(self, element_uid: str, total_since: datetime):
        """
        Update the point in time, the total consumption of a device was reset.

        :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
        :param total_since: Point in time, the total consumption was reset
        """
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices.get(device_uid).consumption_property.get(element_uid).total_since = total_since
        self._logger.debug(f"Updating total since of {element_uid} to {total_since}")
        self._publisher.dispatch(device_uid, (element_uid, total_since))

    def _binary_switch(self, message: dict):
        """ Update a binary switch's state. """
        if message.get("properties").get("property.name") == "state":
            self.update_binary_switch_state(element_uid=message.get("properties").get("uid"),
                                            value=True if message.get("properties").get("property.value.new") == 1
                                            else False)

    def _current_consumption(self, property: dict):
        """ Update current consumption. """
        self.update_consumption(element_uid=property.get("uid"),
                                consumption="current",
                                value=property.get("property.value.new"))

    def _device_change(self, message: dict):
        """ Call method if a new device appears or an old one disappears. """
        if not callable(self.on_device_change):
            self._logger.error("on_device_change is not set.")
            return
        if type(message.get("properties").get("property.value.new")) == list \
           and message.get("properties").get("uid") == "devolo.DevicesPage":
            self.on_device_change(uids=message.get("properties").get("property.value.new"))

    def _device_events(self, message: dict):
        """ If an operation was not successful, we need to correct our internal state. """
        properties = {"properties": message.get("properties").get("property.value.new")}
        properties.get("properties")["uid"] = properties.get("properties").get("widgetElementUID")
        if get_device_type_from_element_uid(properties.get("properties").get("widgetElementUID")) == "devolo.BinarySwitch":
            properties.get("properties")["property.name"] = "state"
            properties.get("properties")["property.value.new"] = int(properties.get("properties").get("data"))
        self.update(properties)

    def _device_online_state(self, message: dict):
        """ Update the device's online state. """
        if message.get("properties").get("property.name") == "status":
            self.update_device_online_state(device_uid=message.get("properties").get("uid"),
                                            value=message.get("properties").get("property.value.new"))

    def _gateway_accessible(self, message: dict):
        """ Update the gateway's state. """
        if message.get("properties").get("property.name") == "gatewayAccessible":
            self.update_gateway_state(accessible=message.get("properties").get("property.value.new").get("accessible"),
                                      online_sync=message.get("properties").get("property.value.new").get("onlineSync"))

    def _meter(self, message: dict):
        """ Update a meter value. """
        property_name = {"currentValue": self._current_consumption,
                         "totalValue": self._total_consumption,
                         "sinceTime": self._since_time}

        property_name[message.get("properties").get("property.name")](message.get("properties"))

    def _since_time(self, property: dict):
        """ Update point in time the total consumption was reset. """
        self.update_total_since(element_uid=property.get("uid"),
                                total_since=property.get("property.value.new"))

    def _total_consumption(self, property: dict):
        """ Update total consumption. """
        self.update_consumption(element_uid=property.get("uid"),
                                consumption="total",
                                value=property.get("property.value.new"))

    def _voltage_multi_level_sensor(self, message: dict):
        """ Update a voltage value. """
        self.update_voltage(element_uid=message.get("properties").get("uid"),
                            value=message.get("properties").get("property.value.new"))



import time

import pytest

from .mocks.mock_websocket import MockWebsocket, MockWebsocketError


@pytest.mark.usefixtures("mprm_instance")
class TestMprmWebsocket:
    def test_get_local_session(self):
        with pytest.raises(NotImplementedError):
            self.mprm.get_local_session()

    def test_get_remote_session(self):
        with pytest.raises(NotImplementedError):
            self.mprm.get_remote_session()

    def test_websocket_connect(self, mock_mprmwebsocket_websocketapp, mprm_session, gateway_instance):
        self.mprm._session = mprm_session
        self.mprm._gateway = gateway_instance
        with pytest.raises(AssertionError):
            self.mprm.websocket_connect()

    def test_websocket_disconnect(self):
        self.mprm._ws = MockWebsocketError()
        with pytest.raises(AssertionError):
            self.mprm.websocket_disconnect()

    def test__on_message(self):
        message = '{"properties": {"com.prosyst.mbs.services.remote.event.sequence.number": 0}}'
        with pytest.raises(NotImplementedError):
            self.mprm._on_message(message)

    def test__on_message_event_sequence(self, mock_mprmwebsocket_on_update):
        event_sequence = self.mprm._event_sequence
        message = '{"properties": {"com.prosyst.mbs.services.remote.event.sequence.number": 5}}'
        self.mprm._on_message(message)
        assert event_sequence != self.mprm._event_sequence
        assert self.mprm._event_sequence == 5

    def test__on_error(self, mock_mprmwebsocket_get_remote_session, mock_mprmwebsocket_websocket_connection):
        self.mprm._ws = MockWebsocket()
        self.mprm._on_error("error")

    def test__try_reconnect(self, mocker, mock_mprmwebsocket_get_local_session_json_decode_error):
        spy = mocker.spy(time, "sleep")
        self.mprm._ws = MockWebsocket()

        self.mprm._local_ip = self.gateway.get("local_ip")
        self.mprm._try_reconnect(0.1)
        spy.assert_called_once_with(0.1)



import pytest

from devolo_home_control_api.homecontrol import get_sub_device_uid_from_element_uid


@pytest.mark.usefixtures("mock_inspect_devices_metering_plug")
@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_mydevolo__call")
class TestHomeControl:

    def test_binary_switch_devices(self):
        assert hasattr(self.homecontrol.binary_switch_devices[0], "binary_switch_property")

    def test_get_publisher(self):
        assert len(self.homecontrol.publisher._events) == 3

    def test_get_sub_device_uid_from_element_uid(self):
        # TODO: Use test data
        assert get_sub_device_uid_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2#2") == 2
        assert get_sub_device_uid_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2") is None

    def test__binary_switch(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[device].binary_switch_property
        assert not hasattr(self.homecontrol.devices.get(device), "binary_switch_property")
        self.homecontrol._binary_switch({"UID": "devolo.BinarySwitch:hdm:ZWave:F6BF9812/2", "properties": {"state": 1}})
        assert hasattr(self.homecontrol.devices.get(device), "binary_switch_property")

    def test__general_device(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._general_device({"UID": "gds.hdm:ZWave:F6BF9812/2",
                                          "properties": {"settings": {"events_enabled": True,
                                                                      "name": self.devices.get("mains").get("itemName"),
                                                                      "zoneId": self.devices.get("mains").get("zone_id"),
                                                                      "icon": self.devices.get("mains").get("icon")}}})
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("general_device_settings"), "events_enabled")

    def test__led(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._led({"UID": "gds.hdm:ZWave:F6BF9812/2", "properties": {"led": True}})
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("led"), "led_setting")

    def test__consumption(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[device].consumption_property
        assert not hasattr(self.homecontrol.devices.get(device), "consumption_property")
        self.homecontrol._meter({"UID": "devolo.Meter:hdm:ZWave:F6BF9812/2", "properties": {
                                 "current": self.devices.get("mains").get("current_consumption"),
                                 "total": self.devices.get("mains").get("total_consumption"),
                                 "sinceTime": self.devices.get("mains").get("properties").get("total_consumption")}})
        assert hasattr(self.homecontrol.devices.get(device), "consumption_property")

    def test__parameter(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._parameter({"UID": "cps.hdm:ZWave:F6BF9812/2",
                                     "properties": {"param_changed": False}})
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("param_changed"), "param_changed")

    def test__protection(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._protection({"UID": "ps.hdm:ZWave:F6BF9812/2",
                                      "properties": {"local_switch": True,
                                                     "remote_switch": False}})
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("protection"), "local_switching")
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("protection"), "remote_switching")

    def test__voltage_multi_level_sensor(self, mock_properties):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[device].voltage_property
        assert not hasattr(self.homecontrol.devices.get(device), "voltage_property")
        self.homecontrol._voltage_multi_level_sensor({"UID": "devolo.VoltageMultiLevelSensor:hdm:ZWave:F6BF9812/2",
                                                      "properties": {
                                                          "current": self.devices.get("mains").get("current_consumption")}})
        assert hasattr(self.homecontrol.devices.get(device), "voltage_property")

    def test_device_change_add(self, mocker, mock_inspect_devices):
        uids = [self.devices.get(device).get("uid") for device in self.devices]
        uids.append("test_uid")
        spy = mocker.spy(self.homecontrol, '_inspect_devices')
        self.homecontrol.device_change(uids)
        spy.assert_called_once_with(["test_uid"])

    def test_device_change_remove(self):
        uids = [self.devices.get(device).get("uid") for device in self.devices]
        del uids[2]
        self.homecontrol.device_change(uids)
        assert self.devices.get("mains").get("uid") not in self.homecontrol.devices.keys()

    @pytest.mark.usefixtures("mock_mprmrest_all_devices")
    @pytest.mark.usefixtures("mock_extract_data_from_element_uids")
    def test__inspect_devices(self, mocker):
        spy = mocker.spy(self.homecontrol, '_inspect_devices')
        self.homecontrol._inspect_devices([self.devices.get("mains")])
        assert spy.call_count == 1



import logging

from ..mydevolo import Mydevolo


class Zwave:
    """
    Representing object for Z-Wave devices.
    """

    def __init__(self, **kwargs):
        self._logger = logging.getLogger(self.__class__.__name__)

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.mydevolo = Mydevolo.get_instance()

        z_wave_info_list = ["href", "manufacturer", "productTypeId", "productId", "name", "brand", "identifier", "isZWavePlus",
                            "deviceType", "zwaveVersion", "specificDeviceClass", "genericDeviceClass"]
        for key in z_wave_info_list:
            setattr(self, key, None)

        self.uid = get_device_uid_from_element_uid(self.elementUIDs[0])

        if self.batteryLevel == -1:
            delattr(self, "batteryLevel")
            delattr(self, "batteryLow")


    def get_property(self, name: str) -> list:
        """
        Get element UIDs to a specified property.

        :param name: Name of the property we want to access
        :return: List of UIDs in this property
        :raises: AttributeError: The property does not exist in this device type
        """
        return [*getattr(self, f"{name}_property").values()]

    def get_zwave_info(self):
        """ Get publicly available information like manufacturer or model. """
        self._logger.debug(f"Getting Z-Wave information for {self.uid}")
        dict = self.mydevolo.get_zwave_products(manufacturer=self.manID,
                                                product_type=self.prodTypeID,
                                                product=self.prodID)
        for key, value in dict.items():
            setattr(self, key, value)

    def is_online(self) -> bool:
        """
        Get the online state of a device.

        :param uid: Device UID, something like hdm:ZWave:CBC56091/24
        :return: False, if device is offline, else True
        """
        return False if self.status == 1 else True


def get_device_type_from_element_uid(element_uid: str) -> str:
    """
    Return the device type of the given element UID.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
    :return: Device type, something like devolo.MultiLevelSensor
    """
    return element_uid.split(":")[0]


def get_device_uid_from_setting_uid(setting_uid: str) -> str:
    """
    Return the device uid of the given setting UID.

    :param setting_uid: Setting UID, something like lis.hdm:ZWave:EB5A9F6C/2
    :return: Device UID, something like hdm:ZWave:EB5A9F6C/2
    """
    setting_uid = setting_uid.split(".", 1)[-1]
    if setting_uid.endswith("secure"):
        return setting_uid.rsplit(":", 1)[0]
    return setting_uid


def get_device_uid_from_element_uid(element_uid: str) -> str:
    """
    Return device UID from the given element UID.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
    :return: Device UID, something like hdm:ZWave:CBC56091/24
    """
    element_uid = element_uid.split(":", 1)[1].split("#")[0]
    if element_uid.endswith("secure"):
        return element_uid.rsplit(":", 1)[0]
    return element_uid



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="devolo_home_control_api",
    version="0.6.4",
    author="Markus Bong, Guido Schmitz",
    author_email="m.bong@famabo.de, guido.schmitz@fedaix.de",
    description="devolo Home Control API in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/2Fake/devolo_home_control_api",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "requests",
        "websocket_client<=0.56.0",
        "zeroconf"
    ],
    python_requires='>=3.6',
)



import pytest

from devolo_home_control_api.backend.mprm_rest import MprmDeviceCommunicationError


@pytest.mark.usefixtures("mprm_instance")
class TestMprmRest:

    def test_get_name_and_element_uids(self, mock_mprmrest__extract_data_from_element_uid, mock_mprmrest__post):
        properties = self.mprm.get_name_and_element_uids("test")
        assert properties == {"itemName": "test_name",
                              "zone": "test_zone",
                              "batteryLevel": "test_battery",
                              "icon": "test_icon",
                              "elementUIDs": "test_element_uids",
                              "settingUIDs": "test_setting_uids",
                              "deviceModelUID": "test_device_model_uid",
                              "status": "test_status"}

    def test_all_devices(self, mock_mprmrest__post):
        devices = self.mprm.all_devices
        assert devices == "deviceUIDs"

    @pytest.mark.usefixtures("mock_response_requests_ReadTimeout")
    def test_post_ReadTimeOut(self, mprm_session, gateway_instance):
        self.mprm._session = mprm_session
        self.mprm._gateway = gateway_instance
        with pytest.raises(MprmDeviceCommunicationError):
            self.mprm.post({"data": "test"})

    @pytest.mark.usefixtures("mock_response_requests_ReadTimeout")
    def test_post_gateway_offline(self, mprm_session, gateway_instance):
        self.mprm._session = mprm_session
        self.mprm._gateway = gateway_instance
        self.mprm._gateway.online = False
        self.mprm._gateway.sync = False
        self.mprm._gateway.local_connection = False
        with pytest.raises(MprmDeviceCommunicationError):
            self.mprm.post({"data": "test"})

    @pytest.mark.usefixtures("mock_response_requests_invalid_id")
    def test_post_invalid_id(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._data_id = 0
        with pytest.raises(ValueError):
            self.mprm.post({"data": "test"})

    @pytest.mark.usefixtures("mock_response_requests_valid")
    def test_post_valid(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._data_id = 1
        assert self.mprm.post({"data": "test"}).get("id") == 2

    def test_get_data_from_uid_list(self, mock_mprmrest__post):
        properties = self.mprm.get_data_from_uid_list(["test"])
        assert properties[0].get("properties").get("itemName") == "test_name"



import pytest

from devolo_home_control_api.properties.property import WrongElementError
from devolo_home_control_api.properties.voltage_property import VoltageProperty


@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_mprmrest__extract_data_from_element_uid")
class TestVoltageProperty:
    def test_voltage_property_invalid(self, gateway_instance, mprm_session):
        with pytest.raises(WrongElementError):
            VoltageProperty(gateway=gateway_instance, session=mprm_session, element_uid="invalid", current=0.0)

    def test_fetch_voltage_valid(self):
        voltage = self.homecontrol.devices.get(self.devices.get("mains").get("uid")).\
            voltage_property.get(f"devolo.VoltageMultiLevelSensor:{self.devices.get('mains').get('uid')}").fetch_voltage()
        assert voltage == 237



from typing import Any

from requests import Session

from ..devices.gateway import Gateway
from .property import Property, WrongElementError


class SettingsProperty(Property):
    """
    Object for settings. It stores the different settings.

    :param gateway: Instance of a Gateway object
    :param session: Instance of a requests.Session object
    :param element_uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24#2
    :param **kwargs: Any setting, that shall be available in this object
    """

    def __init__(self, gateway: Gateway, session: Session, element_uid: str, **kwargs: Any):
        if element_uid.split(".")[0] not in ["lis", "gds", "cps", "ps"]:
            raise WrongElementError()

        super().__init__(gateway=gateway, session=session, element_uid=element_uid)
        self.setting_uid = element_uid
        for key, value in kwargs.items():
            setattr(self, key, value)


    def fetch_general_device_settings(self) -> tuple:
        """
        Update and return the events enabled setting. If a device shall report to the diary, this is true.

        :return: Events enabled or not
        """
        response = self.get_data_from_uid_list([self.setting_uid])
        self.name = response.get("properties").get("name")
        self.icon = response.get("properties").get("icon")
        self.zone_id = response.get("properties").get("zoneID")
        self.events_enabled = response.get("properties").get("eventsEnabled")
        return self.name, self.icon, self.zone_id, self.events_enabled

    def fetch_led_setting(self) -> bool:
        """
        Update and return the led setting.

        :return: LED setting
        """
        response = self.get_data_from_uid_list([self.setting_uid])
        self.led_setting = response.get("properties").get("led")
        return self.led_setting

    def fetch_param_changed_setting(self) -> bool:
        """
        Update and return the param changed setting.

        :param setting_uid: Settings UID to look at. Usually starts with cps.hdm.
        :return: True, if parameter was changed
        """
        response = self.get_data_from_uid_list([self.setting_uid])
        self.param_changed = response.get("properties").get("paramChanged")
        return self.param_changed

    def fetch_protection_setting(self, protection_setting: str) -> bool:
        """
        Update and return the protection setting. There are only two protection settings: local and remote switching.

        :param protection_setting: Look at local or remote switching.
        :return: Switching is protected or not
        """
        if protection_setting not in ["local", "remote"]:
            raise ValueError("Only local and remote are possible protection settings")
        response = self.get_data_from_uid_list([self.setting_uid])
        if protection_setting == "local":
            self.local_switching = response.get("properties").get("localSwitch")
            return self.local_switching
        else:
            self.remote_switching = response.get("properties").get("remoteSwitch")
            return self.remote_switching



import pytest

from devolo_home_control_api.properties.consumption_property import ConsumptionProperty
from devolo_home_control_api.properties.property import WrongElementError


@pytest.mark.usefixtures("home_control_instance")
class TestConsumption:
    def test_consumption_property_invalid(self, gateway_instance, mprm_session):
        with pytest.raises(WrongElementError):
            ConsumptionProperty(gateway=gateway_instance,
                                session=mprm_session,
                                element_uid="invalid",
                                current=0.0,
                                total=0.0,
                                total_since=1231412412)

    def test_fetch_consumption_invalid(self):
        with pytest.raises(ValueError):
            self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
                .consumption_property.get(self.devices.get("mains").get("elementUIDs")[0]).fetch_consumption("invalid")

    def test_fetch_consumption_valid(self, mock_mprmrest__extract_data_from_element_uid):
        current = self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .consumption_property.get(self.devices.get("mains").get("properties").get("elementUIDs")[0])\
            .fetch_consumption(consumption_type="current")
        total = self.homecontrol.devices.get(self.devices.get("mains").get("uid"))\
            .consumption_property.get(self.devices.get("mains").get("properties").get("elementUIDs")[0])\
            .fetch_consumption(consumption_type="total")
        assert current == 0.58
        assert total == 125.68



import threading
from typing import Optional

import requests

from .backend.mprm import Mprm
from .devices.gateway import Gateway
from .devices.zwave import Zwave, get_device_type_from_element_uid, get_device_uid_from_setting_uid, \
    get_device_uid_from_element_uid
from .properties.binary_switch_property import BinarySwitchProperty
from .properties.consumption_property import ConsumptionProperty
from .properties.settings_property import SettingsProperty
from .properties.voltage_property import VoltageProperty
from .publisher.publisher import Publisher
from .publisher.updater import Updater


class HomeControl(Mprm):
    """
    Representing object for your Home Control setup. This is more or less the glue between your devolo Home Control Central
    Unit, your devices and their properties.

    :param gateway_id: Gateway ID (aka serial number), typically found on the label of the device
    :param url: URL of the mPRM (typically leave it at default)
    """

    def __init__(self, gateway_id: str, url: str = "https://homecontrol.mydevolo.com"):
        self._gateway = Gateway(gateway_id)
        self._session = requests.Session()
        self._session.url = url

        super().__init__()

        # Create the initial device dict
        self.devices = {}
        self._inspect_devices(self.all_devices)

        self.device_names = dict(zip([self.devices.get(device).itemName for device in self.devices],
                                     [self.devices.get(device).uid for device in self.devices]))

        self.publisher = Publisher([device for device in self.devices])

        self.updater = Updater(devices=self.devices, gateway=self._gateway, publisher=self.publisher)
        self.updater.on_device_change = self.device_change

        threading.Thread(target=self.websocket_connect).start()


    @property
    def binary_switch_devices(self) -> list:
        """ Get all binary switch devices. """
        return [self.devices.get(uid) for uid in self.devices if hasattr(self.devices.get(uid), "binary_switch_property")]


    def device_change(self, device_uids: list):
        """
        React on new devices or removed devices. As the Z-Wave controller can only be in inclusion or exclusion mode, we
        assume, that you cannot add and remove devices at the same time. So if the number of devices increases, there is
        a new one and if the number decreases, a device was removed.

        :param device_uids: List of UIDs known by the backend
        """
        if len(device_uids) > len(self.devices):
            new_device = [device for device in device_uids if device not in self.devices]
            self._logger.debug(f"New device found {new_device}")
            self._inspect_devices([new_device[0]])
            self.publisher = Publisher([device for device in self.devices])
        else:
            devices = [device for device in self.devices if device not in device_uids]
            self._logger.debug(f"Device {devices} removed")
            del self.devices[devices[0]]
        self.updater.devices = self.devices

    def on_update(self, message):
        """
        Initialize steps needed to update properties on a new message.

        :param message: Message because of which we need to update properties
        """
        self.updater.update(message)


    def _binary_switch(self, uid_info: dict):
        """ Process BinarySwitch properties. """
        if not hasattr(self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))], "binary_switch_property"):
            self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))].binary_switch_property = {}
        self._logger.debug(f"Adding binary switch property to {get_device_uid_from_element_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))].binary_switch_property[uid_info.get("UID")] = \
            BinarySwitchProperty(session=self._session,
                                 gateway=self._gateway,
                                 element_uid=uid_info.get("UID"),
                                 state=True if uid_info.get("properties").get("state") == 1 else False)

    def _general_device(self, uid_info: dict):
        """ Process general device setting (gds) properties. """
        self._logger.debug(f"Adding general device settings to {get_device_uid_from_setting_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_setting_uid(uid_info.get('UID'))]. \
            settings_property["general_device_settings"] = \
            SettingsProperty(session=self._session,
                             gateway=self._gateway,
                             element_uid=uid_info.get("UID"),
                             events_enabled=uid_info.get("properties").get("settings").get("eventsEnabled"),
                             name=uid_info.get("properties").get("settings").get("name"),
                             zone_id=uid_info.get("properties").get("settings").get("zoneID"),
                             icon=uid_info.get("properties").get("settings").get("icon"))

    def _inspect_devices(self, devices: list):
        """ Inspect device properties of given list of devices. """
        devices_properties = self.get_data_from_uid_list(devices)
        for device_properties in devices_properties:
            properties = device_properties.get("properties")
            self.devices[device_properties.get("UID")] = Zwave(**properties)
            self.devices[device_properties.get("UID")].settings_property = {}
            threading.Thread(target=self.devices[device_properties.get("UID")].get_zwave_info).start()

        elements = {"devolo.BinarySwitch": self._binary_switch,
                    "devolo.Meter": self._meter,
                    "devolo.VoltageMultiLevelSensor": self._voltage_multi_level_sensor,
                    "lis.hdm": self._led,
                    "gds.hdm": self._general_device,
                    "cps.hdm": self._parameter,
                    "ps.hdm": self._protection
                    }
        # List comprehension gets the list of uids from every device
        nested_uids_lists = [(uid.get("properties").get('settingUIDs')
                              + uid.get("properties").get("elementUIDs"))
                             for uid in devices_properties]

        # List comprehension gets all uids into one list to make one big call against the mPRM
        uid_list = [uid for sublist in nested_uids_lists for uid in sublist]

        for uid_info in self.get_data_from_uid_list(uid_list):
            if uid_info.get("UID") is not None:
                elements.get(get_device_type_from_element_uid(uid_info.get("UID")), self._unknown)(uid_info)

    def _led(self, uid_info: dict):
        """ Process LED information setting (lis) properties. """
        self._logger.debug(f"Adding led settings to {get_device_uid_from_setting_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_setting_uid(uid_info.get('UID'))].settings_property["led"] = \
            SettingsProperty(session=self._session,
                             gateway=self._gateway,
                             element_uid=uid_info.get("UID"),
                             led_setting=uid_info.get("properties").get("led"))

    def _meter(self, uid_info: dict):
        """ Process Meter properties. """
        if not hasattr(self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))], "consumption_property"):
            self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))].consumption_property = {}
        self._logger.debug(f"Adding consumption property to {get_device_uid_from_element_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))].consumption_property[uid_info.get("UID")] = \
            ConsumptionProperty(session=self._session,
                                gateway=self._gateway,
                                element_uid=uid_info.get("UID"),
                                current=uid_info.get("properties").get("currentValue"),
                                total=uid_info.get("properties").get("totalValue"),
                                total_since=uid_info.get("properties").get("sinceTime"))

    def _parameter(self, uid_info: dict):
        """ Process custom parameter setting (cps) properties."""
        self._logger.debug(f"Adding parameter settings to {get_device_uid_from_setting_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_setting_uid(uid_info.get('UID'))].settings_property["param_changed"] = \
            SettingsProperty(session=self._session,
                             gateway=self._gateway,
                             element_uid=uid_info.get('UID'),
                             param_changed=uid_info.get('properties').get("paramChanged"))

    def _protection(self, uid_info: dict):
        """ Process protection setting (ps) properties. """
        self._logger.debug(f"Adding protection settings to {get_device_uid_from_setting_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_setting_uid(uid_info.get('UID'))].settings_property["protection"] = \
            SettingsProperty(session=self._session,
                             gateway=self._gateway,
                             element_uid=uid_info.get('UID'),
                             local_switching=uid_info.get("properties").get("localSwitch"),
                             remote_switching=uid_info.get("properties").get("remoteSwitch"))

    def _unknown(self, uid_info: dict):
        """ Ignore unknown properties. """
        self._logger.debug(f"Found an unexpected element uid: {uid_info.get('UID')}")

    def _voltage_multi_level_sensor(self, uid_info: dict):
        """ Process VoltageMultiLevelSensor properties. """
        if not hasattr(self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))], "voltage_property"):
            self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))].voltage_property = {}
        self._logger.debug(f"Adding voltage property to {get_device_uid_from_element_uid(uid_info.get('UID'))}.")
        self.devices[get_device_uid_from_element_uid(uid_info.get("UID"))].voltage_property[uid_info.get("UID")] = \
            VoltageProperty(session=self._session,
                            gateway=self._gateway,
                            element_uid=uid_info.get("UID"),
                            current=uid_info.get("properties").get("value"))


def get_sub_device_uid_from_element_uid(element_uid: str) -> Optional[int]:
    """
    Return the sub device uid of the given element UID.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
    :return: Sub device UID, something like 2
    """
    return None if "#" not in element_uid else int(element_uid.split("#")[-1])



import pytest
from datetime import datetime


@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_publisher_dispatch")
class TestUpdater:
    def test_update_binary_switch_state_valid(self, fill_device_data):
        uid = self.devices.get('mains').get("uid")
        binary_switch_property = self.homecontrol.devices.get(uid).binary_switch_property
        state = binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state
        self.homecontrol.updater.update_binary_switch_state(element_uid=f"devolo.BinarySwitch:{uid}",
                                                            value=True)
        assert state != binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state

    def test_update_binary_switch_state_group(self, fill_device_data):
        self.homecontrol.updater.update_binary_switch_state(element_uid=f"devolo.BinarySwitch:devolo.smartGroup.1",
                                                            value=True)
        # If the updater does something with this uid it will raise a KeyError.
        # If this happens the test will fail although there is no assert

    def test_update_consumption_valid(self, fill_device_data):
        uid = self.devices.get('mains').get("uid")
        consumption_property = self.homecontrol.devices.get(uid).consumption_property
        current_before = consumption_property.get(f"devolo.Meter:{uid}").current
        total_before = consumption_property.get(f"devolo.Meter:{uid}").total
        self.homecontrol.updater.update_consumption(element_uid=f"devolo.Meter:{uid}",
                                                    consumption="current", value=1.58)
        self.homecontrol.updater.update_consumption(element_uid=f"devolo.Meter:{uid}",
                                                    consumption="total", value=254)
        assert current_before != consumption_property.get(f"devolo.Meter:{uid}").current
        assert total_before != consumption_property.get(f"devolo.Meter:{uid}").total
        assert consumption_property.get(f"devolo.Meter:{uid}").current == 1.58
        assert consumption_property.get(f"devolo.Meter:{uid}").total == 254

    def test_device_online_state(self):
        uid = self.devices.get('mains').get("uid")
        online_state = self.homecontrol.devices.get(uid).status
        self.homecontrol.updater.update_device_online_state(device_uid=self.devices.get('mains').get('uid'),
                                                            value=1)
        assert self.homecontrol.devices.get(uid).status == 1
        assert self.homecontrol.devices.get(uid).status != online_state

    def test_update_total_since(self, fill_device_data):
        element_uid = self.devices.get('mains').get("elementUIDs")[0]
        total_since = self.homecontrol.devices.get(self.devices.get('mains').get("uid"))\
            .consumption_property.get(element_uid).total_since
        now = datetime.now()
        self.homecontrol.updater.update_total_since(element_uid=element_uid, total_since=now)
        assert total_since != self.homecontrol.devices.get(self.devices.get('mains').get("uid"))\
            .consumption_property.get(element_uid).total_since
        assert self.homecontrol.devices.get(self.devices.get('mains').get("uid"))\
                   .consumption_property.get(element_uid).total_since == now

    def test_update_gateway_state(self):
        self.homecontrol.updater.update_gateway_state(accessible=True, online_sync=False)
        assert self.homecontrol._gateway.online
        assert not self.homecontrol._gateway.sync

    def test_update_voltage_valid(self, fill_device_data):
        uid = self.devices.get('mains').get("uid")
        voltage_property = self.homecontrol.devices.get(uid).voltage_property
        current_voltage = \
            voltage_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").current
        self.homecontrol.updater.update_voltage(element_uid=f"devolo.VoltageMultiLevelSensor:{uid}",
                                                value=257)
        assert current_voltage != \
            voltage_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").current
        assert voltage_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").current == 257

    def test_update(self):
        self.homecontrol.updater.update(message={"properties":
                                        {"uid": f"devolo.BinarySwitch:{self.devices.get('mains').get('uid')}"}})

    def test_update_invalid(self):
        self.homecontrol.updater.update(message={"properties":
                                        {"uid": "fibaro"}})

    def test__binary_switch(self):
        uid = self.devices.get('mains').get("uid")
        self.homecontrol.devices.get(uid).binary_switch_property \
            .get(f"devolo.BinarySwitch:{uid}").state = True
        state = self.homecontrol.devices.get(uid).binary_switch_property \
            .get(f"devolo.BinarySwitch:{uid}").state
        self.homecontrol.updater._binary_switch(message={"properties":
                                                {"property.name": "state",
                                                 "uid": f"devolo.BinarySwitch:{uid}",
                                                 "property.value.new": 0}})
        state_new = self.homecontrol.devices.get(uid).binary_switch_property \
            .get(f"devolo.BinarySwitch:{uid}").state
        assert state != state_new

    def test__device_change(self):
        self.homecontrol.updater.on_device_change = lambda uids: AssertionError()
        try:
            self.homecontrol.updater._device_change(message={"properties":
                                                    {"uid": "devolo.DevicesPage",
                                                     "property.value.new": []}})
            assert False
        except AssertionError:
            assert True

    def test__device_change_error(self, mocker):
        self.homecontrol.updater.on_device_change = None
        spy = mocker.spy(self.homecontrol.updater._logger, "error")
        self.homecontrol.updater._device_change({})
        spy.assert_called_once_with("on_device_change is not set.")

    def test__device_online_state(self):
        uid = self.devices.get('mains').get("uid")
        online_state = self.homecontrol.devices.get(uid).status
        self.homecontrol.updater._device_online_state(message={"properties": {"uid": uid,
                                                                              "property.name": "status",
                                                                              "property.value.new": 1}})
        assert self.homecontrol.devices.get(uid).status == 1
        assert online_state != self.homecontrol.devices.get(uid).status

    def test__device_events(self):
        uid = self.devices.get('mains').get("uid")
        self.homecontrol.devices.get(uid).binary_switch_property \
            .get(f"devolo.BinarySwitch:{uid}").state = True
        self.homecontrol.updater._device_events(message={"properties":
                                                         {"property.value.new":
                                                          {"widgetElementUID": f"devolo.BinarySwitch:{uid}",
                                                           "property.name": "state",
                                                           "data": 0}}})
        assert not self.homecontrol.devices.get(uid).binary_switch_property .get(f"devolo.BinarySwitch:{uid}").state


    def test__gateway_accessible(self):
        self.homecontrol._gateway.online = True
        self.homecontrol._gateway.sync = True
        accessible = self.homecontrol._gateway.online
        online_sync = self.homecontrol._gateway.sync
        self.homecontrol.updater._gateway_accessible(message={"properties": {"property.name": "gatewayAccessible",
                                                                             "property.value.new": {"accessible": False,
                                                                                                    "onlineSync": False}}})
        accessible_new = self.homecontrol._gateway.online
        online_sync_new = self.homecontrol._gateway.sync
        assert accessible != accessible_new
        assert online_sync != online_sync_new

    def test__meter(self):
        uid = self.devices.get('mains').get("uid")
        self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").current = 5
        self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").total = 230
        total = self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").total
        # Changing current value
        self.homecontrol.updater._meter(message={"properties": {"property.name": "currentValue",
                                                                "uid": f"devolo.Meter:{uid}",
                                                                "property.value.new": 7}})
        current_new = self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").current
        # Check if current value has changed
        assert current_new == 7
        # Check if total has not changed
        assert total == self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").total
        # Changing total value
        self.homecontrol.updater._meter(message={"properties": {"property.name": "totalValue",
                                                                "uid": f"devolo.Meter:{uid}",
                                                                "property.value.new": 235}})
        total_new = self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").total
        # Check if total value has changed
        assert total_new == 235
        # Check if current value has not changed
        assert self.homecontrol.devices.get(uid).consumption_property \
            .get(f"devolo.Meter:{uid}").current == current_new

    def test__voltage_multi_level_sensor(self):
        uid = self.devices.get('mains').get("uid")
        self.homecontrol.devices.get(uid).voltage_property \
            .get(f"devolo.VoltageMultiLevelSensor:{uid}").current = 231
        current = self.homecontrol.devices.get(uid).voltage_property \
            .get(f"devolo.VoltageMultiLevelSensor:{uid}").current
        self.homecontrol.updater._voltage_multi_level_sensor(message={"properties":
                                                             {"uid": f"devolo.VoltageMultiLevelSensor:{uid}",
                                                              "property.value.new": 234}})
        current_new = self.homecontrol.devices.get(uid).voltage_property \
            .get(f"devolo.VoltageMultiLevelSensor:{uid}").current

        assert current_new == 234
        assert current != current_new


    def test__since_time(self):
        now = datetime.now()
        total_since = self.homecontrol.devices['hdm:ZWave:F6BF9812/2'] \
            .consumption_property['devolo.Meter:hdm:ZWave:F6BF9812/2'].total_since
        self.homecontrol.updater._since_time({"uid": "devolo.Meter:hdm:ZWave:F6BF9812/2",
                                              "property.value.new": now})
        new_total_since = self.homecontrol.devices['hdm:ZWave:F6BF9812/2'] \
            .consumption_property['devolo.Meter:hdm:ZWave:F6BF9812/2'].total_since
        assert total_since != new_total_since
        assert new_total_since == now
